package seal

import(
	"context"
	"fmt"
	"io"
	"io/ioutil"
	"os"
	"math/rand"
	"path/filepath"
	"time"

	"golang.org/x/xerrors"
	"github.com/mitchellh/go-homedir"
	"github.com/minio/blake2b-simd"
	commcid "github.com/filecoin-project/go-fil-commcid"
	"github.com/filecoin-project/go-state-types/abi"
	
	spproof "fil_integrate/build/proof"
	"fil_integrate/build"
	"fil_integrate/build/storage"
	"fil_integrate/build/storiface"
	// "fil_integrate/extern/sector-storage/ffiwrapper"
	"fil_integrate/seal/basicfs"
)

func TestAggregateWindowPoSt(
	sectorSize abi.SectorSize,
	numSectors int,
	numAggregate int,
) error {
	sdir, err := homedir.Expand("~/tmp/bench")
	if err != nil {
		return err
	}

	err = os.MkdirAll(sdir, 0775) //nolint:gosec
	if err != nil {
		return xerrors.Errorf("creating sectorbuilder dir: %w", err)
	}

	tsdir, err := ioutil.TempDir(sdir, "bench")
	if err != nil {
		return err
	}
	defer func() {
		if err := os.RemoveAll(tsdir); err != nil {
			log.Warn("remove all: ", err)
		}
	}()

	// TODO: pretty sure this isnt even needed?
	if err := os.MkdirAll(tsdir, 0775); err != nil && !os.IsExist(err){
		return err
	}
	sbfs := &basicfs.Provider{
			Root: tsdir,
		}
	sb ,err := New(sbfs)
	if err != nil{
		return err
	}

	file := rand.New(rand.NewSource(1587))
	trand := blake2b.Sum256([]byte("ticket-preimage"))
	ticket := abi.SealRandomness(trand[:])
	ctx := context.TODO()
	var challenge [32]byte
	rand.Read(challenge[:])

	var randomnesses []abi.PoStRandomness
	var sealedSectorsinfo [][]spproof.SectorInfo
	var sectorCount []uint
	var proofs []spproof.PoStProof
	sealProofType := spt(sectorSize)

	start := time.Now()
	for i := 0; i < numAggregate; i++{
		for j := 0; j < numSectors; j++{
			var pieces []abi.PieceInfo

			sid := storage.SectorRef{
				ID: abi.SectorID{
					Miner:  1000,
					Number: abi.SectorNumber(i*numSectors+j),
				},
				ProofType: sealProofType,
			}

			piece, err := sb.AddPiece(ctx, sid, nil, abi.PaddedPieceSize(sectorSize).Unpadded(), file)
			if err != nil {
				return err
			}
			pieces = append(pieces, piece)

			pc1out, err := sb.SealPreCommit1(ctx, sid, ticket, pieces)
			if err != nil {
				return xerrors.Errorf("commit: %w", err)
			}

			cids, err := sb.SealPreCommit2(ctx, sid, pc1out)
			if err != nil {
				return xerrors.Errorf("commit: %w", err)
			}

			comm_r := cids.Sealed
			sb.PutCommRIntoDir(sid.ID, comm_r)
		}
	}
	log.Infof("Sealed [%d] Sectors Done", numSectors*numAggregate)
	sealed := time.Now()

	for i := 0; i < numAggregate; i++{
		var sealedSectors []spproof.SectorInfo
		for j := 0; j < numSectors; j++{
			sectorID := abi.SectorID{
				Miner:  1000,
				Number: abi.SectorNumber(i*numSectors+j),
			}
			comm_r, err := sb.GetCommRFromDir(sectorID)
			if err != nil {
				return err
			}
			sealedSectors = append(sealedSectors, spproof.SectorInfo{
				SealedCID:    comm_r,
				SectorNumber: sectorID.Number,
				SealProof:    sealProofType,
			})
		}
		sealedSectorsinfo = append(sealedSectorsinfo, sealedSectors)
	}
	log.Infof("Read [%d] Commitment Rplication Done", numSectors*numAggregate)
	loadCommr := time.Now()

	for i := 0; i < numAggregate; i++{
		log.Infof("[%d] Generating Window-Post", i)
		proof, _, err := sb.GenerateWindowPoStProofs(ctx, 1000, sealedSectorsinfo[i], challenge[:])
		if err != nil {
			return err
		}

		proofs = append(proofs, proof...)
		randomnesses = append(randomnesses, challenge[:])
		sectorCount = append(sectorCount, uint(numSectors))
	}
	log.Infof("Generate [%d] Window-PoSt Done", numAggregate)
	genWindowPoSt := time.Now()

	aggregateProof1, err := sb.AggregateWindowPoStProofs(spproof.AggregateWindowPostInfos{
			AggregateType: DefaultAggregationType(),
			Randomnesses:  randomnesses, 
			SectorCount:   sectorCount,
		}, proofs)
	if err != nil {
		return err
	}

	aggregateProofsCold := time.Now()

	aggregateProof2, err := sb.AggregateWindowPoStProofs(spproof.AggregateWindowPostInfos{
			AggregateType: DefaultAggregationType(),
			Randomnesses:  randomnesses, 
			SectorCount:   sectorCount,
		}, proofs)
	if err != nil {
		return err
	}
	aggregateProofsHot := time.Now()

	poStType, _ := sealProofType.RegisteredWindowPoStProof()
	svi1 := spproof.AggregateWindowPostInfos{
			PoStType:         poStType,
			AggregateType:    DefaultAggregationType(),
			Miner:            abi.ActorID(1000),
			AggregationProof: aggregateProof1,
			Randomnesses:     randomnesses,      
		}
	svi2 := spproof.AggregateWindowPostInfos{
			PoStType:         poStType,
			AggregateType:    DefaultAggregationType(),
			Miner:            abi.ActorID(1000),
			AggregationProof: aggregateProof2,
			Randomnesses:     randomnesses,      
		}

	ok, err := ProofVerifier.VerifyAggregateWindowPostProofs(svi1, sealedSectorsinfo)
	if err != nil {
		return err
	}
	if ok {
		fmt.Println("Aggregated proof is true")
	} else{
		fmt.Println("Aggregated proof is false")
	}
	verifyProofsCold := time.Now()

	ok, err = ProofVerifier.VerifyAggregateWindowPostProofs(svi2, sealedSectorsinfo)
	if err != nil {
		return err
	}
	if ok {
		fmt.Println("Aggregated proof is true")
	} else{
		fmt.Println("Aggregated proof is false")
	}
	verifyProofsHot := time.Now()

	fmt.Printf("Seal %d sectors using %s\n", numSectors*numAggregate, sealed.Sub(start))
	fmt.Printf("Read %d comm_r using %s\n", numAggregate*numSectors, loadCommr.Sub(sealed))
	fmt.Printf("Generate %d window-post using %s\n", numAggregate, genWindowPoSt.Sub(loadCommr))
	fmt.Printf("Aggregate %d window-post Proofs(cold) using %s\n", numAggregate, aggregateProofsCold.Sub(genWindowPoSt))
	fmt.Printf("Aggregate %d window-post Proofs(hot) using %s\n", numAggregate, aggregateProofsHot.Sub(aggregateProofsCold))
	fmt.Printf("Verify Aggregation Window-PoSt Proofs(cold) using %s\n", verifyProofsCold.Sub(aggregateProofsHot))
	fmt.Printf("Verify Aggregation Window-PoSt Proofs(hot) using %s\n", verifyProofsHot.Sub(verifyProofsCold))

	return nil
}

func TestSealAndUnseal() error {
	//********************need (sb,ctx,sid,sectorSize,file,seed,ticket,challenge)****************//
	sdir, err := homedir.Expand("~/tmp/bench")
	if err != nil {
		return err
	}

	err = os.MkdirAll(sdir, 0775) //nolint:gosec
	if err != nil {
		return xerrors.Errorf("creating sectorbuilder dir: %w", err)
	}

	tsdir, err := ioutil.TempDir(sdir, "bench")
	if err != nil {
		return err
	}
	defer func() {
		if err := os.RemoveAll(tsdir); err != nil {
			log.Warn("remove all: ", err)
		}
	}()

	// TODO: pretty sure this isnt even needed?
	if err := os.MkdirAll(tsdir, 0775); err != nil {
		return err
	}
	sbfs := &basicfs.Provider{
			Root: tsdir,
		}
	sb ,err := New(sbfs)
	if err != nil{
		return err
	}
	ctx := context.TODO()
	sectorSize := abi.SectorSize(8*1024*1024)
	sid := storage.SectorRef{
		ID: abi.SectorID{
			Miner:  1000,
			Number: 0,
		},
		ProofType: spt(sectorSize),
	}
	file := rand.New(rand.NewSource(1587))
	seed := []byte{1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 255}
	trand := blake2b.Sum256([]byte("ticket-preimage"))
	ticket := abi.SealRandomness(trand[:])
	var challenge [32]byte
	rand.Read(challenge[:])
	//ADD PIECES
	var existingPieceSizes []abi.UnpaddedPieceSize
	var pieces []abi.PieceInfo
	var sealedSectors []spproof.SectorInfo

	piece, err := sb.AddPiece(ctx, sid, existingPieceSizes, abi.PaddedPieceSize(sectorSize/2).Unpadded(), file)
	if err != nil {
		return err
	}

	existingPieceSizes = append(existingPieceSizes, piece.Size.Unpadded())
	pieces = append(pieces, piece)

	piece, err = sb.AddPiece(ctx, sid, existingPieceSizes, abi.PaddedPieceSize(sectorSize/2).Unpadded(), file)
	if err != nil {
		return err
	}

	existingPieceSizes = append(existingPieceSizes, piece.Size.Unpadded())
	pieces = append(pieces, piece)

	//SEAL
	cids, proof1, err := sb.Sealed(ctx, sid, seed, ticket, pieces)
	if err != nil {
		return err
	}

	sealedSectors = append(sealedSectors, spproof.SectorInfo{
		SealedCID:    cids.Sealed,
		SectorNumber: sid.ID.Number,
		SealProof:    sid.ProofType,
	})

	proof2, err := sb.GenerateCommit2Proof(ctx, sid, seed, ticket, pieces, cids)
	if err != nil {
		return err
	}

	ok, err := sb.CheckPieceAndDataRoot(sid, cids.Unsealed, pieces)
	if err != nil {
		return err
	}

	if !ok {
		return xerrors.Errorf("commd and pieces info don't match")
	}

	//verify proof
	svi1 := spproof.SealVerifyInfo{
		SectorID:              sid.ID,
		SealedCID:             cids.Sealed,
		SealProof:             sid.ProofType,
		Proof:                 proof1,
		DealIDs:               nil,
		Randomness:            ticket,
		InteractiveRandomness: seed,
		UnsealedCID:           cids.Unsealed,
	}

	svi2 := spproof.SealVerifyInfo{
		SectorID:              sid.ID,
		SealedCID:             cids.Sealed,
		SealProof:             sid.ProofType,
		Proof:                 proof2,
		DealIDs:               nil,
		Randomness:            ticket,
		InteractiveRandomness: seed,
		UnsealedCID:           cids.Unsealed,
	}

	ok, err = ProofVerifier.VerifySeal(svi1)
	if err != nil {
		return err
	}
	if !ok {
		return xerrors.Errorf("porep proof for sector %d was invalid", sid.ID.Number)
	}

	ok, err = ProofVerifier.VerifySeal(svi2)
	if err != nil {
		return err
	}
	if !ok {
		return xerrors.Errorf("porep proof for sector %d was invalid", sid.ID.Number)
	}

	proof, _, err := sb.GenerateWindowPoStProofs(ctx, sid.ID.Miner, sealedSectors, challenge[:])

	wpvi := spproof.WindowPoStVerifyInfo{
		Randomness:        challenge[:],
		Proofs:            proof,
		ChallengedSectors: sealedSectors,
		Prover:            sid.ID.Miner,
	}

	ok, err = ProofVerifier.VerifyWindowPoSt(wpvi)
	if err != nil {
		return err
	}
	if !ok {
		log.Error("window post verification failed")
	}

	return nil

}

func TestSplitDataInToPieces() error {
	sdir, err := homedir.Expand("~/tmp/bench")
	if err != nil {
		return err
	}

	err = os.MkdirAll(sdir, 0775) //nolint:gosec
	if err != nil {
		return xerrors.Errorf("creating sectorbuilder dir: %w", err)
	}

	tsdir, err := ioutil.TempDir(sdir, "bench")
	if err != nil {
		return err
	}
	// defer func() {
	// 	if err := os.RemoveAll(tsdir); err != nil {
	// 		log.Warn("remove all: ", err)
	// 	}
	// }()

	// TODO: pretty sure this isnt even needed?
	if err := os.MkdirAll(tsdir, 0775); err != nil {
		return err
	}
	sbfs := &basicfs.Provider{
			Root: tsdir,
		}
	sb, err := New(sbfs)
	if err != nil{
		return err
	}
	ctx := context.TODO()
	sectorSize := abi.SectorSize(8*1024*1024)

	root, err := homedir.Expand("~/tmp")
	if err != nil {
		return err
	}
	filename := filepath.Join(root, "input.dat")
	err = generateRandomData(filename)
	if err != nil {
		return err
	}
	in, err := os.OpenFile(filename, os.O_RDONLY, 0644)
	if err != nil {
		return err
	}
	defer in.Close()

	finalHash, piecesHash, err := sb.EncodeDataToPieces(ctx, sectorSize, in)
	if err != nil{
		return err
	}

	for i, pieceHash := range(piecesHash) {
		sid := storage.SectorRef{
			ID: abi.SectorID{
				Miner:  1000,
				Number: abi.SectorNumber(i),
			},
			ProofType: spt(sectorSize),
		}

		filename = filepath.Join(tsdir, "pieces", fmt.Sprintf("%x.dat", pieceHash[:]))
		f, err := os.OpenFile(filename, os.O_RDONLY|os.O_CREATE, 0644)
		if err != nil {
			return err
		}
		defer f.Close()
		piece, err := sb.AddPiece(ctx, sid, nil, abi.PaddedPieceSize(sectorSize).Unpadded(), f)
		if err != nil {
			return err
		}
		commp, err := commcid.CIDToPieceCommitmentV1(piece.PieceCID)
		if err != nil {
			return err
		}
		if string(commp[:]) != string(pieceHash[:]) {
			fmt.Printf("commp and piece hash mismatch, %x != %x\n", commp[:], pieceHash[:])
		} else {
			fmt.Printf("commp and piece hash match, %x == %x\n", commp[:], pieceHash[:])
		}
	}

	filename = filepath.Join(root, "output.dat")
	out, err := os.OpenFile(filename, os.O_WRONLY|os.O_CREATE, 0644)
	if err != nil {
		return err
	}
	defer out.Close()

	err = decodePiecesToData(sb, ctx, tsdir, sectorSize, finalHash, out)
	if err != nil {
		return err
	}

	ok, err := checkDecodedFile(root)
	if err != nil {
		return err
	}
	if !ok {
		fmt.Println("decode pieces failed")
	} else {
		fmt.Println("decode pieces success")
	}

	return nil
}

func Test() int {
	var buf1 []byte
	var buf2 []byte
	buf1 = append(buf1, 0,1,2,3)
	buf2 = append(buf2, 10,20,30,40)
	buf1 = append(buf2, buf1...)
	fmt.Println(buf1, len(buf1), buf1[4])
	fmt.Println(buf2, len(buf2), buf2[4])
	return 0
}

func spt(ssize abi.SectorSize) abi.RegisteredSealProof {
	spt, err := build.SealProofTypeFromSectorSize(ssize, NewestNetworkVersion)
	if err != nil {
		panic(err)
	}

	return spt
}

func generateRandomData(filename string) error {
	Datasize := 128*1024*1024
	buf := make([]byte, Datasize)
	b := []byte("random string!")
	for i:=0; i<Datasize; i=i+32{
		temphash := blake2b.Sum256(b)
		copy(b, temphash[:])
		copy(buf[i:i+32],temphash[:])
	}
	_ = os.Remove(filename)
	f,err := os.OpenFile(filename, os.O_CREATE|os.O_WRONLY, 0644)
	if err != nil {
		return err
	}
	defer f.Close()
	_, err = f.Write(buf[:])
	if err != nil {
		return err
	}
	return nil
}

func decodePiecesToData(sb *Sealer, ctx context.Context, tsdir string, sectorSize abi.SectorSize, finalHash storage.Hash, out io.Writer) error {
	// var piecesHash []storage.Hash
	DataLen := abi.PaddedPieceSize(sectorSize).Unpadded() - 8
	filename := filepath.Join(tsdir, "pieces", fmt.Sprintf("%x.dat", finalHash[:]))
	fmt.Printf("Decode: %x.dat\n", finalHash[:])
	file, err := os.OpenFile(filename, os.O_RDONLY, 0644)
	if err != nil {
		return err
	}
	defer file.Close()

	// hasPre, preHash, Data, commData, err := sb.DecodePiece(ctx, sectorSize, file, 0, storiface.UnpaddedByteIndex(DataLen))
	data, err := sb.DecodePiece(ctx, sectorSize, file, 0, storiface.UnpaddedByteIndex(DataLen))
	if err != nil {
		return err
	}
	buf := data.Data[:]
	piecesHash := data.PieceHash

	for ; data.HasPre; {
		filename = filepath.Join(tsdir, "pieces", fmt.Sprintf("%x.dat", data.PreHash[:]))
		fmt.Printf("Decode: %x.dat\n", data.PreHash[:])
		file, err := os.OpenFile(filename, os.O_RDONLY, 0644)
		if err != nil {
			return err
		}
		defer file.Close()

		// hasPre, preHash, Data, hashData, err = sb.DecodePiece(ctx, sectorSize, file, 0, storiface.UnpaddedByteIndex(DataLen))
		data, err = sb.DecodePiece(ctx, sectorSize, file, 0, storiface.UnpaddedByteIndex(DataLen))
		if err != nil{
			return err
		}
		piecesHash = append(data.PieceHash, piecesHash...)
	}

	for _, pieceHash := range piecesHash {
		filename = filepath.Join(tsdir, "pieces", fmt.Sprintf("%x.dat", pieceHash[:]))
		fmt.Printf("Decode: %x.dat\n", pieceHash[:])
		file, err := os.OpenFile(filename, os.O_RDONLY, 0644)
		data, err := sb.DecodePiece(ctx, sectorSize, file, 0, storiface.UnpaddedByteIndex(DataLen))
		if err != nil {
			return err
		}
		for wbuf := data.Data[:]; len(wbuf) > 0; {
			n, err := out.Write(wbuf)
			if err != nil{
				return err
			}
			wbuf = wbuf[n:]
		}
	}
	for wbuf := buf[:]; len(wbuf) > 0; {
		n, err := out.Write(wbuf)
		if err != nil{
			return err
		}
		wbuf = wbuf[n:]
	}
	return nil
}

func checkDecodedFile(root string) (bool, error) {
	filename := filepath.Join(root, "input.dat")
	in, err := os.Open(filename)
	if err != nil {
		return false, err
	}
	defer in.Close()

	filename = filepath.Join(root, "output.dat")
	out, err := os.Open(filename)
	if err != nil {
		return false, err
	}
	defer out.Close()

	inBuf := make([]byte, 2<<20)
	outBuf := make([]byte, 2<<20)

	for{
		readin, inerr := in.Read(inBuf[:])
		if err != nil && err != io.EOF{
			return false, err
		}

		readout, outerr := out.Read(outBuf[:])
		if err != nil && err != io.EOF{
			return false, err
		}

		if readin != readout {
			return false, xerrors.Errorf("the output data and input data do not match")
		}

		for index := 0; index < readin; index++ {
			if inBuf[index] != outBuf[index] {
				return false, xerrors.Errorf("the output data and input data do not match at: %d input is %u, output is %u",index,inBuf[index],outBuf[index])
			}
		}

		if inerr == io.EOF && outerr == io.EOF {
			break
		}
	}
	return true, nil
}